;*************************************
; M62 BIOS:  8255LCD.asm - 8255 PPI LCD Driver
;*************************************
; Supports '4 rows x 20 columns' parallel LCD panel.
; Displays "LCD Ready", then ">" prompt, and any ASCII key typed. It will not stream BIOS responses; a larger display is needed.
; Wraps each row at 20 characters.
; Wraps rows contiquously: 1,2,3,4 (not default of 1,3,2,4).
;************************************

.ORG	$4000

ROM_START	.EQU 0000h
ROM_END		.EQU 3FFFh				; 16KB ROM.
RAM_START	.EQU 4000h
RAM_END		.EQU 7FFFh				; 16KB RAM.
; This config is necessary to use /Base_2 as MEB.

;*********************
; 8255 Config
;*********************
BasePPI .EQU 	$38 			; Default 82C55A PPI address is $00.
pAadr	.EQU	BasePPI + $00	; Port A address. Output.
pBadr	.EQU	BasePPI + $01	; Port B address. Output.
pCadr	.EQU	BasePPI + $02	; Port C address. Input/output.
CWRadr 	.EQU	BasePPI + $03	; PPI Control Word Register.
CWRcfg	.EQU	$80				; 1 00 0 0 00 0 
								; Port mode, *A output, *C-upper output, C-lower output, B output.

;*********************
; 8255 Port Pins
;*********************
; PA[0..6]	LCD Data Out		Port A: to LCD DB0-DB6. 
; PA[7]		LCD Busy In			Port A: to LCD DB7.
; PC[5..7]	LCD Control Out		Port C-upper: to LCD RS, RW, E. 

;*********************
; LCD Functions
; See http://zed80.com/Z80-RETRO/Resources/Boards/00-M62/11-M62_20x4_LCD_Bd/8255-CWR-PortC-Bit-Modifications.jpg
;*********************
lcdCommand 	.EQU 0ah		; Reset LCD bit #4   portC bit #5(RS) 101-0
lcdData 	.EQU 0bh		;   Set LCD bit #4   portC bit #5(RS) 101-1
lcdWrite 	.EQU 0ch		; Reset LCD bit #5   portC bit #6(RW) 110-0
lcdRead 	.EQU 0dh		;   Set LCD bit #5   portC bit #6(RW) 110-1
lcdDisable	.EQU 0eh		; Reset LCD bit #6   portC bit #7(E)  111-0
lcdEnable 	.EQU 0fh		;   Set LCD bit #6   portC bit #7(E)  111-1 

Positions 	.EQU 20			; Number of LCD columns (20x2 and 20x4).
numCmnd		.EQU 04h		; Number of Commands.
numData 	.EQU 0Bh		; Number of data bytes in string.

;********************
; LCD Display Commands:
;********************
lcdClrCmnd:		.EQU	$01		; 01 is Clear command to clear the display and home the cursor.
lcdHomeCmnd:	.EQU	$02		; 02 is Home command: cursor is homed.
lcdRow1Cmnd:	.EQU	$80		; LCD 80h + 0. 		 First row starts with address 80h. Combined address is 80h + 00h = 80h.
lcdRow2Cmnd:	.EQU	$C0		; LCD 80h + address. Second row starts with address 40h. Combined address is 80h + 40h = C0h.
lcdRow3Cmnd:	.EQU	$94		; LCD 80h + address. Third row starts with address 14h. Combined address is 80h + 14h = 94h.
lcdRow4Cmnd:	.EQU	$D4		; LCD 80h + address. Fourth row starts with address 54h. Combined address is 80h + 54h = D4h.




START_LCD:					;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;*********************
; Init RAM and 8255:	
;*********************
		LD SP, RAM_END		; Top of RAM.
		LD A, CWRcfg		; Set CWR (0x80) for all 3 ports as outputs: A out, C-upper out (also B out and C-lower out).
		OUT (CWRadr), A		; OUT (CWRadr), CWRcfg. CWRadr is 0x38 as per Peter's design.
		LD E, 0h			; Clear register E to count column position.

;*********************	
; Init the LCD functional parameters:
;*********************
; Set RS to Command (low), set RW to Write (low), and output to CWR in pin mode.
; Set LCD config to 38h (Clear, 8-bit, 2-row, 5x7 pixel characters) and Output to portA. 
; Pulse (high then low) the lcdEnable line for the LCD to read portChigh pins RW & RS, and portA data lines DB7-DB0. 
		LD	HL,	DELAY_LCD		; 1.6ms. LCD initialization time.
		CALL	loopDELAY	
initLCD: 
		CALL lcdWriteCMND		; Reset RS pin (low) for command, Reset RW pin (low) for writing
		LD A, 38h				; LCD function set Command: 8-bit mode, AC=0, 2 rows, 5x7 characters.
		OUT (pAadr), A			; Execute the LCD 38h Command.
		CALL pulseLCD			; After RW and RS line have been set/reset and a command/data has been sent to the LCD,
								; pulse the LCD (high then low) to read the aforementioned parameters.
;*********************
; Init the LCD display parameters: 
;*********************
; Clear display ($01), set DD RAM address to Row 1 ($80), turn on display with cursor hidden ($0C), and set entry mode 6 ($06).
		LD HL, cmndBeginAdr		; Set HL to point to the first Command.
		LD B, numCmnd			; Number of Commands (4) to be sent is put in counter B.
nextCmnd: 
		LD A,	(HL)			
		CALL lcdWriteCMNDStr	; Send contents of (HL) as a Command.
		INC HL					; Point to the next Command.
		DJNZ nextCmnd			; Decrement B. Loop until all 4 commands are sent.

;***************************************************
; Send the string "LCD Ready":
;***************************************************
		LD HL, dataBeginAdr		; Set HL to point the first string byte.
		LD B, numData			; Put the number of string bytes to be sent in B.
nextData:
		LD A,	(HL)			
		CALL lcdWriteDATAStr	; Send (HL) as string data. 
		INC HL					; Point to the next string byte.
		DJNZ nextData			; Loop until all string bytes are sent.
fadeOutLCD:	
;		LD	HL,	DELAY_FULL64K
;		CALL	loopDELAY
;		LD	HL,	DELAY_FULL64K
;		CALL	loopDELAY
		call	longDELAY		;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
		CALL	lcdClear
		CALL	lcdPrompt		; '>' BIOS prompt.
		JP	0100h				; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Return from app to BIOS.
		;RET					
		
;***************************************************
; Write the entire DATA string one byte at a time. Requires knowing how many bytes.
;***************************************************
lcdWriteDATAStr:
		CALL lcdWriteDATA
		OUT (pAadr), A	
		PUSH BC		
		PUSH DE	
		CALL pulseLCD
		POP DE
		POP BC
		RET	
		
lcdWriteDATA:
		LD C, CWRadr	
		LD D, lcdData		; Data to follow.
		OUT (C), D			; Set RS pin (high) high for data.
		LD D,lcdWrite		
		OUT (C), D			; Reset RW pin (low) for writing to LCD.
		RET	
		
;*********************
; Write the entire COMMAND string one byte at a time. Requires knowing how many bytes.
;*********************
lcdWriteCMNDStr:
		CALL lcdWriteCMND	; Uses C and D registers
		OUT (pAadr), A		; Place Command to be written into portA.
		PUSH BC				; B is used for commands counter, C is used for CWRcfg.
		PUSH DE				; D is used for RS/RW pins, E is used for commmand or data.
		CALL pulseLCD		; C & D
		POP DE				; Restore DE.
		POP BC				; Restore BC.
		RET		
	
lcdWriteCMND:
		LD C, CWRadr			 
		LD D, lcdCommand	; Command to follow. 
		OUT (C), D			; Reset RS pin (low) for command/instruction.
		LD D,lcdWrite		; Write command.
		OUT (C), D			; Reset RW pin (low) for writing to LCD.
		RET
	
pulseLCD:					; After RW and RS line have been set/reset and a command/data has been sent to the LCD, pulse the LCD on then off to read the aforementioned.
		LD D,	lcdEnable	; Start LCD pulse.
		OUT (C), D			; ......Enable the LCD. 
		PUSH HL				; ......
		LD	HL,	DELAY_LCD   ; ......
		CALL	loopDELAY	; ......Slow the pulse.
		POP HL              ; ......
		LD D,	lcdDisable	; ......Disable the LCD.
		OUT (C), D			; End LCD pulse.
		RET
	
lcdClear:  					; Clear the LCD screen and home the cursor.
		PUSH AF
		LD	A,	lcdClrCmnd	; $01
		CALL	lcdWriteCMNDStr
		POP	AF
		RET
			
lcdHome: 					; Home the cursor to the beginning of the row.
		PUSH AF
		LD A,	lcdHomeCmnd		; $02
		CALL	lcdWriteCMNDStr
		POP	AF
		RET
		
lcdPrompt:	
		PUSH AF				; Display BIOS command prompt.
		LD	A,	'>'
		CALL lcdWriteDATAStr
		POP	AF
		RET
		
lcdRow1:					; Home the cursor to the beginning of the first row.
		PUSH AF
		LD A,	lcdRow1Cmnd		; $80
		CALL	lcdWriteCMNDStr
		POP	AF
		JP Retrn1
		
lcdRow2:					; Home the cursor to the beginning of the second row.
		PUSH AF
		LD A,	lcdRow2Cmnd		; $C0
		CALL	lcdWriteCMNDStr
		POP	AF
		JP Retrn1
	
lcdRow3:					; Home the cursor to the beginning of the third row.
		PUSH AF
		LD A,	lcdRow3Cmnd		; $94
		CALL	lcdWriteCMNDStr
		POP	AF
		JP Retrn1
	
lcdRow4:					; Home the cursor to the beginning of the fourth row.
		PUSH AF
		LD A,	lcdRow4Cmnd		; $D4
		CALL	lcdWriteCMNDStr	
		POP	AF
		JP Retrn1

;lcdCRLF:	
;		CP $0D					; CR = $0D
;		JP	Z,	Retrn1
;		CP $0A					; LF = $0A
;		JP	Z,	Retrn1
;		CALL	lcdWriteDATAStr
;		CALL	lcdRow1
;		RET
;*******************************************************
; lcdRowSelect
;*******************************************************
; Checks if CR or LF are sent and skips them: they are weird characters on a 20x4 LCD panel.	
; Panel row wrap is 1 to 3 to 2 to 4 from factory. Changed it to wrap contiguously 1,2,3,4.
; Each user character is displayed contiguously until all 4 lines have wrapped, then screen 
; is cleared and homed to 0,0.
; '>' BIOS prompt is always at 0,0.
;*******************************************************
lcdRowSelect:
		CP $0D					; CR = $0D
		JP	Z,	Retrn1
		CP $0A					; LF = $0A
		JP	Z,	Retrn1
		CALL	lcdWriteDATAStr
		INC	B
		PUSH	AF
		LD A,	B
		CP	20
		JP Z,	lcdRow2
		CP	40
		JP Z,	lcdRow3
		CP	60
		JP Z,	lcdRow4
		CP	80
		JP Z,	ResetB
Retrn1:	POP	AF	
		RET
ResetB:	POP	AF
		LD	B,	0
		CALL lcdClear
		CALL lcdPrompt
		RET

DELAY_LCD:		.EQU	$0220	; 1.6ms. LCD initialization time.
DELAY_FULL64K:	.EQU	$FFFF	; 360ms
;********************
; Delay routine
;********************
loopDELAY:  				
		NOP
		NOP
        DEC L
        JP NZ, loopDELAY
        DEC H
        JP NZ, loopDELAY
        RET            

    
;moveRight1:				; Moves the cursor one column to the right of the current position.
;		PUSH AF
;mR1:	
;		CALL lcdWriteCMND
;		LD A, E				; Positions/columns.
;		CP 20				; Have we reached 20th column in a 2x20 or 4x20 LCD panel?
;		JP Z, Position0		; Yes, reset the position to first column of second row.
;		LD A, 0C0h			
;		ADD A, E			; No, add the position value to the start of the second row (c0h).
;		OUT (pAadr), A		; Send the command to position the cursor.
;		CALL pulseLCD
;		INC E 				; Increment position for next key to be pressed and displayed.
;		POP AF
;		RET
;	
;Position0:					; We have reached the 20th position/column. Reset to the zeroeth position.
;		LD E, 0h
;		LD A, 0h
;		CALL lcdClear		; Clear the screen.
;		CALL lcdRow2		; Home the cursor to the beginning of the second row.
;		JP mR1

    
cmndBeginAdr:				; Clear display, set DD RAM adress, turn on display with cursor hidden, set entry mode.
		.db 01h,80h,0ch,06h  	
	
dataBeginAdr:				; "> LCD Ready"
		.db 3Eh,20h,4Ch,43h,44h,20,52h,65h,61h,64h,79h
		
#INCLUDE	"longDelays.asm"	; Delays several seconds long.
		
.END